import bz2
import gzip
import re
import zipfile
from io import BytesIO
from pathlib import Path
from typing import Union

from .hatanaka import crx2rnx, rnx2crx

try:
    # unlzw is implemented in C and more efficient that Python-based unlzw3.
    # Much more difficult to install, though.
    from unlzw import unlzw
except ImportError:
    from unlzw3 import unlzw

__all__ = [
    'decompress', 'decompress_on_disk', 'get_decompressed_path',
    'compress', 'compress_on_disk', 'get_compressed_path'
]


def decompress(content: Union[Path, str, bytes], *,
               skip_strange_epochs: bool = False) -> bytes:
    """Decompress compressed RINEX files.

    Any RINEX files compressed with Hatanaka compression (.crx|.##d) and/or with a conventional
    compression format (.gz|.Z|.zip|.bz2) are decompressed to their plain RINEX counterpart.
    Already decompressed files are returned as is.

    Compression type is deduced automatically from the file contents.
    File extensions are not validated (as is done in decompress_on_disk()).

    Parameters
    ----------
    content : Path or str or bytes
        Path to a compressed RINEX file or file contents as a bytes object.
    skip_strange_epochs : bool, default False
        For Hatanaka decompression.
        Warn and skip strange epochs instead of raising an exception.
        This option may be used for salvaging usable data when middle of the Compact
        RINEX file is missing. The data after the missing part, are, however, useless
        until the compression operation of all data are initialized at some epoch.
        Using this together with of reinit_every_nth option of rnx2crx may be effective.
        Caution: It is assumed that no change in the list of data types happens in the
        lost part of the data.

    Returns
    -------
    bytes
        Decompressed RINEX file contents.

    Raises
    ------
    HatanakaException
        On any errors during Hatanaka decompression.
    ValueError
        For invalid file contents.
    """
    if isinstance(content, (Path, str)):
        content = _decompress(Path(content).read_bytes(), skip_strange_epochs)[1]
    elif not isinstance(content, bytes):
        raise ValueError('input must be either a path or a binary string')
    return _decompress(content, skip_strange_epochs)[1]


def decompress_on_disk(path: Union[Path, str], *, skip_strange_epochs: bool = False) -> Path:
    """Decompress compressed RINEX files and write the resulting file to disk.

    Any RINEX files compressed with Hatanaka compression (.crx|.##d) and/or with a conventional
    compression format (.gz|.Z|.zip|.bz2) are decompressed to their plain RINEX counterpart.
    Already decompressed files are ignored.

    Compression type is deduced automatically from the file contents.
    Input file extensions are loosely validated to assign a correct extension
    to the decompressed file.

    Parameters
    ----------
    path : Path or str
        Path to a compressed RINEX file.
    skip_strange_epochs : bool, default False
        For Hatanaka decompression.
        Warn and skip strange epochs instead of raising an exception.
        This option may be used for salvaging usable data when middle of the Compact
        RINEX file is missing. The data after the missing part, are, however, useless
        until the compression operation of all data are initialized at some epoch.
        Using this together with of reinit_every_nth option of rnx2crx may be effective.
        Caution: It is assumed that no change in the list of data types happens in the
        lost part of the data.

    Returns
    -------
    Path
        Path to the decompressed RINEX file.

    Raises
    ------
    HatanakaException
        On any errors during Hatanaka decompression.
    ValueError
        For invalid file contents.
    """
    path = Path(path)
    is_obs, txt = _decompress(path.read_bytes(), skip_strange_epochs=skip_strange_epochs)
    out_path = get_decompressed_path(path, is_obs)
    if out_path == path:
        # file does not need decompressing
        return out_path
    with out_path.open('wb') as f_out:
        f_out.write(txt)
    return out_path


def get_decompressed_path(path: Union[Path, str], is_obs: bool) -> Path:
    """Get the decompressed path corresponding to a compressed RINEX file after decompression.

    Parameters
    ----------
    path : path or str
        Path to the compression RINEX file.
    is_obs : bool
        Whether the RINEX file contains observation data.
        Needed for correct renaming of Hatanaka-compressed files.

    Returns
    -------
    Path
        The path of the resulting RINEX file after decompression.
    """
    path = Path(path)
    parts = path.name.split('.')
    if len(parts) <= 1:
        raise ValueError(f"'{str(path)}' is not a valid RINEX file name")
    if parts[-1].lower() in ['z', 'gz', 'bz2', 'zip']:
        parts.pop()
    suffix = parts[-1]
    if is_obs:
        is_valid = re.match(r'^(?:crx|rnx|\d\d[od])$', suffix, flags=re.I)
        if not is_valid:
            raise ValueError(f"'{str(path)}' is not a valid RINEX file name")
        if suffix[2] == 'd':
            suffix = suffix[:2] + 'o'
        elif suffix[2] == 'D':
            suffix = suffix[:2] + 'O'
        elif suffix == 'crx':
            suffix = 'rnx'
        elif suffix == 'CRX':
            suffix = 'RNX'
    out_path = path.parent / '.'.join(parts[:-1] + [suffix])
    return out_path


def compress(content: Union[Path, str, bytes], *, compression: str = 'gz',
             skip_strange_epochs: bool = False,
             reinit_every_nth: int = None) -> bytes:
    """Compress RINEX files.

    Applies Hatanaka (if observation data) and optionally a conventional compression (gzip by default)
    to the provided RINEX file. Already compressed files are returned as is.

    Parameters
    ----------
    content : Path or str or bytes
        Path to a RINEX file or file contents as a bytes object.
    compression : 'gz' (default), 'bz2', or 'none'
        Which compression (if any) to apply in addition to the Hatanaka compression.
    skip_strange_epochs : bool, default False
        For Hatanaka compression. Warn and skip strange epochs instead of raising an exception.
    reinit_every_nth : int, optional
        For Hatanaka compression. Initialize the compression operation at every # epochs.
        When some part of the Compact RINEX file is lost, the data can not be recovered
        thereafter until all the data arc are initialized for differential operation.
        This option may be used to increase chances to recover parts of data by using the
        skip_strange option of crx2rnx at the cost of increasing the file size.

    Returns
    -------
    bytes
        Compressed RINEX file contents.

    Raises
    ------
    HatanakaException
        On any errors during Hatanaka compression.
    ValueError
        For invalid file contents.
    """
    if isinstance(content, (Path, str)):
        content = Path(content).read_bytes()
    elif not isinstance(content, bytes):
        raise ValueError('input must be either a path or a binary string')
    return _compress(content, compression, skip_strange_epochs, reinit_every_nth)[1]


def compress_on_disk(path: Union[Path, str], *, compression: str = 'gz',
                     skip_strange_epochs: bool = False,
                     reinit_every_nth: int = None) -> Path:
    """Compress RINEX files.

    Applies Hatanaka (if observation data) and optionally a conventional compression (gzip by default)
    to the provided RINEX file. Already compressed files are ignored.

    Parameters
    ----------
    path : Path or str
        Path to a RINEX file.
    compression : 'gz' (default), 'bz2', or 'none'
        Which compression (if any) to apply in addition to the Hatanaka compression.
    skip_strange_epochs : bool, default False
        For Hatanaka compression. Warn and skip strange epochs instead of raising an exception.
    reinit_every_nth : int, optional
        For Hatanaka compression. Initialize the compression operation at every # epochs.
        When some part of the Compact RINEX file is lost, the data can not be recovered
        thereafter until all the data arc are initialized for differential operation.
        This option may be used to increase chances to recover parts of data by using the
        skip_strange option of crx2rnx at the cost of increasing the file size.

    Returns
    -------
    Path
        Path to the compressed RINEX file.

    Raises
    ------
    HatanakaException
        On any errors during Hatanaka compression.
    ValueError
        For invalid file contents.
    """
    path = Path(path)
    if path.name.lower().endswith(('.gz', '.bz2', '.z', '.zip')):
        # already compressed
        return path
    is_obs, txt = _compress(path.read_bytes(), compression=compression,
                            skip_strange_epochs=skip_strange_epochs,
                            reinit_every_nth=reinit_every_nth)
    out_path = get_compressed_path(path, is_obs, compression)
    out_path.write_bytes(txt)
    return out_path


def get_compressed_path(path, is_obs, compression='gz'):
    """Get the compressed path corresponding to a RINEX file after compression.

    Parameters
    ----------
    path : path or str
        Path to the compression RINEX file.
    is_obs : bool
        Whether the RINEX file contains observation data.
        Needed for correct renaming of Hatanaka-compressed files.
    compression : 'gz' (default), 'bz2', or 'none'
        Compression (if any) applied in addition to the Hatanaka compression.

    Returns
    -------
    Path
        The path of the resulting RINEX file after compression.
    """
    path = Path(path)
    parts = path.name.split('.')
    if len(parts) < 2:
        raise ValueError(f"'{str(path)}' is not a valid RINEX file name")
    suffix = parts[-1]
    if is_obs:
        is_valid = re.match(r'^(?:crx|rnx|\d\d[od])$', suffix, flags=re.I)
        if not is_valid:
            raise ValueError(f"'{str(path)}' is not a valid RINEX file name")
        if suffix[2] == 'o':
            suffix = suffix[:2] + 'd'
        elif suffix[2] == 'O':
            suffix = suffix[:2] + 'D'
        elif suffix == 'rnx':
            suffix = 'crx'
        elif suffix == 'RNX':
            suffix = 'CRX'
    out_parts = parts[:-1] + [suffix]
    if compression != 'none':
        out_parts.append(compression)
    out_path = path.parent / '.'.join(out_parts)
    return out_path


def _is_lzw(magic_bytes: bytes) -> bool:
    return magic_bytes == b'\x1F\x9D'


def _is_gz(magic_bytes: bytes) -> bool:
    return magic_bytes == b'\x1F\x8B'


def _is_zip(magic_bytes: bytes) -> bool:
    return magic_bytes == b'\x50\x4B'


def _is_bz2(magic_bytes: bytes) -> bool:
    return magic_bytes == b'\x42\x5A'


def _decompress(txt: bytes, skip_strange_epochs: bool) -> (bool, bytes):
    if len(txt) < 2:
        raise ValueError('empty file')
    magic_bytes = txt[:2]

    if _is_gz(magic_bytes):
        return _decompress_hatanaka(gzip.decompress(txt), skip_strange_epochs)
    if _is_bz2(magic_bytes):
        return _decompress_hatanaka(bz2.decompress(txt), skip_strange_epochs)
    elif _is_zip(magic_bytes):
        with zipfile.ZipFile(BytesIO(txt), 'r') as z:
            flist = z.namelist()
            if len(flist) == 0:
                raise ValueError('zip archive is empty')
            elif len(flist) > 1:
                raise ValueError('more than one file in zip archive')
            with z.open(flist[0], 'r') as f:
                return _decompress_hatanaka(f.read(), skip_strange_epochs)
    elif _is_lzw(magic_bytes):
        return _decompress_hatanaka(unlzw(txt), skip_strange_epochs)
    else:
        return _decompress_hatanaka(txt, skip_strange_epochs)


def _decompress_hatanaka(txt: bytes, skip_strange_epochs) -> (bool, bytes):
    if len(txt) < 80:
        raise ValueError('file is too short to be a valid RINEX file')

    is_crinex = b'COMPACT RINEX' in txt[:80]
    if is_crinex:
        txt = crx2rnx(txt, skip_strange_epochs=skip_strange_epochs)
    is_obs = b'OBSERVATION DATA' in txt[:80]
    return is_obs, txt


def _compress(txt: bytes, compression, skip_strange_epochs, reinit_every_nth) -> (bool, bytes):
    is_obs, txt = _compress_hatanaka(txt, skip_strange_epochs, reinit_every_nth)
    if compression == 'gz':
        return is_obs, gzip.compress(txt)
    elif compression == 'bz2':
        return is_obs, bz2.compress(txt)
    elif compression == 'zip':
        raise NotImplementedError('zip compression is not supported')
    elif compression == 'none':
        return is_obs, txt
    else:
        raise ValueError(f"invalid compression '{compression}'")


def _compress_hatanaka(txt: bytes, skip_strange_epochs, reinit_every_nth) -> (bool, bytes):
    if len(txt) < 80:
        raise ValueError('file is too short to be a valid RINEX file')

    is_obs = b'OBSERVATION DATA' in txt[:80]
    if is_obs:
        return is_obs, rnx2crx(txt, skip_strange_epochs=skip_strange_epochs,
                               reinit_every_nth=reinit_every_nth)
    else:
        is_obs = b'COMPACT RINEX' in txt[:80]
        return is_obs, txt



import io
import shutil
import sys
from contextlib import contextmanager

import pytest

from hatanaka.cli import compress_cli, decompress, decompress_cli
from .conftest import clean, get_data_path
from .test_general_compression import compress_pairs, decompress_pairs


@contextmanager
def replace_stdin(content):
    orig = sys.stdin
    target = io.BytesIO()
    target.write(content)
    target.seek(0)
    target.buffer = target
    sys.stdin = target
    yield
    sys.stdin = orig


@contextmanager
def replace_stdout():
    orig = sys.stdout
    target = io.BytesIO()
    sys.stdout = target
    target.buffer = target
    yield target
    target.seek(0)
    sys.stdout = orig


@pytest.mark.parametrize(
    'input_suffix, expected_suffix',
    decompress_pairs
)
def test_decompress_cli(tmp_path, crx_sample, rnx_str, input_suffix, expected_suffix):
    # prepare
    in_file = 'sample' + input_suffix
    sample_path = tmp_path / in_file
    shutil.copy(get_data_path(in_file), sample_path)
    # decompress
    retcode = decompress_cli([str(sample_path)])
    # check
    assert retcode == 0
    expected_path = tmp_path / ('sample' + expected_suffix)
    assert sample_path.exists()
    assert expected_path.exists()
    assert clean(expected_path.read_text()) == clean(rnx_str)


@pytest.mark.parametrize(
    'input_suffix, compression, expected_suffix',
    compress_pairs
)
def test_compress_cli(tmp_path, crx_sample, rnx_bytes, input_suffix, compression,
                      expected_suffix):
    # prepare
    in_file = 'sample' + input_suffix
    sample_path = tmp_path / in_file
    shutil.copy(get_data_path(in_file), sample_path)
    # compress
    retcode = compress_cli([
        str(sample_path),
        '--compression', compression
    ])
    # check
    assert retcode == 0
    expected_path = tmp_path / ('sample' + expected_suffix)
    assert sample_path.exists()
    assert expected_path.exists()
    assert clean(decompress(expected_path)) == clean(rnx_bytes)


def test_decompress_cli_stdin(rnx_bytes):
    in_data = get_data_path('sample.crx.gz').read_bytes()
    with replace_stdin(in_data), replace_stdout() as stdout:
        decompress_cli([])
    assert clean(stdout.getvalue()) == clean(rnx_bytes)


def test_compress_cli_stdin(rnx_bytes):
    in_data = get_data_path('sample.rnx').read_bytes()
    with replace_stdin(in_data), replace_stdout() as stdout:
        compress_cli([])
    result = stdout.getvalue()
    assert clean(result) != clean(rnx_bytes)
    assert clean(decompress(result)) == clean(rnx_bytes)


def test_cli_delete(tmp_path, rnx_bytes):
    # prepare
    in_file = 'sample.rnx'
    sample_path = tmp_path / in_file
    shutil.copy(get_data_path(in_file), sample_path)
    # compress
    retcode = compress_cli([str(sample_path), '--delete'])
    # check
    assert retcode == 0
    expected_path = tmp_path / 'sample.crx.gz'
    assert not sample_path.exists()
    assert expected_path.exists()
    assert clean(decompress(expected_path)) == clean(rnx_bytes)


def test_cli_no_delete_on_warnings(tmp_path, rnx_bytes):
    # prepare
    in_file = 'sample.rnx'
    sample_path = tmp_path / in_file
    txt = get_data_path(in_file).read_bytes()
    sample_path.write_bytes(txt + b'\0\0\0')
    # compress
    with pytest.warns(UserWarning) as record:
        retcode = compress_cli([str(sample_path), '--delete'])
    # check
    assert len(record) == 1
    assert record[0].message.args[0].startswith('rnx2crx: null characters')
    assert retcode == 2
    expected_path = tmp_path / 'sample.crx.gz'
    assert sample_path.exists()
    assert expected_path.exists()
    assert clean(decompress(expected_path)) == clean(rnx_bytes)


def test_cli_no_delete_unchanged(tmp_path, rnx_str):
    # prepare
    in_file = 'sample.crx.gz'
    sample_path = tmp_path / in_file
    shutil.copy(get_data_path(in_file), sample_path)
    # compress
    retcode = compress_cli([str(sample_path), '--delete'])
    # check
    assert retcode == 0
    expected_path = tmp_path / 'sample.crx.gz'
    assert sample_path.exists()
    assert expected_path == sample_path



import re

import pytest
from importlib_resources import files

import hatanaka.test.data


def get_data_path(fname):
    for x in ['.21o', '.rnx']:
        fname = fname.replace(x.lower(), '.rnx')
        fname = fname.replace(x.upper(), '.rnx')
    for x in ['.21d', '.crx']:
        fname = fname.replace(x.lower(), '.crx')
        fname = fname.replace(x.upper(), '.crx')
    return files(hatanaka.test.data) / fname


def clean(txt):
    # Remove rnx2crx version number and timestamp for testing
    pattern = '^RNX2CRX.+CRINEX PROG / DATE$'
    if isinstance(txt, bytes):
        return re.sub(pattern.encode(), b'', txt.replace(b'\r', b''), flags=re.M)
    return re.sub(pattern, '', txt.replace('\r', ''), flags=re.M)


@pytest.fixture
def crx_sample():
    return get_data_path('sample.crx')


@pytest.fixture
def rnx_sample():
    return get_data_path('sample.rnx')


@pytest.fixture
def crx_str(crx_sample):
    return crx_sample.read_text()


@pytest.fixture
def rnx_str(rnx_sample):
    return rnx_sample.read_text()


@pytest.fixture
def crx_str_stream(crx_sample):
    with crx_sample.open('r', encoding='ascii') as f:
        yield f


@pytest.fixture
def rnx_str_stream(rnx_sample):
    with rnx_sample.open('r', encoding='ascii') as f:
        yield f


@pytest.fixture
def crx_bytes(crx_sample):
    return crx_sample.read_bytes()


@pytest.fixture
def rnx_bytes(rnx_sample):
    return rnx_sample.read_bytes()


@pytest.fixture
def crx_bytes_stream(crx_sample):
    with crx_sample.open('rb') as f:
        yield f


@pytest.fixture
def rnx_bytes_stream(rnx_sample):
    with rnx_sample.open('rb') as f:
        yield f



import gzip
import io
import shutil

import pytest

from hatanaka import compress, compress_on_disk, decompress, decompress_on_disk
from .conftest import clean, get_data_path

decompress_pairs = [
    ('.crx', '.rnx'),
    ('.CRX', '.RNX'),
    ('.21d', '.21o'),
    ('.21D', '.21O'),
    ('.21d.gz', '.21o'),
    ('.21d.Z', '.21o'),
    ('.21d.bz2', '.21o'),
    ('.21d.zip', '.21o'),
    ('.crx.gz', '.rnx'),
    ('.crx.Z', '.rnx'),
    ('.crx.bz2', '.rnx'),
    ('.crx.zip', '.rnx'),
    ('.rnx', '.rnx'),
    ('.rnx.gz', '.rnx'),
    ('.rnx.Z', '.rnx'),
    ('.rnx.zip', '.rnx'),
    ('.rnx.bz2', '.rnx'),
    ('.21o', '.21o'),
    ('.21o.gz', '.21o'),
]


@pytest.mark.parametrize(
    'input_suffix, expected_suffix',
    decompress_pairs
)
def test_decompress(tmp_path, crx_sample, rnx_bytes, input_suffix, expected_suffix):
    # prepare
    sample_path = tmp_path / ('sample' + input_suffix)
    in_file = 'sample' + input_suffix
    shutil.copy(get_data_path(in_file), sample_path)
    # decompress
    converted = decompress(sample_path)
    # check
    assert clean(converted) == clean(rnx_bytes)
    converted = decompress(sample_path.read_bytes())
    assert clean(converted) == clean(rnx_bytes)


@pytest.mark.parametrize(
    'input_suffix, expected_suffix',
    decompress_pairs
)
def test_decompress_on_disk(tmp_path, crx_sample, rnx_bytes, input_suffix, expected_suffix):
    # prepare
    sample_path = tmp_path / ('sample' + input_suffix)
    in_file = 'sample' + input_suffix
    shutil.copy(get_data_path(in_file), sample_path)
    # decompress
    out_path = decompress_on_disk(sample_path)
    # check
    assert out_path.exists()
    assert out_path == tmp_path / ('sample' + expected_suffix)
    assert clean(out_path.read_bytes()) == clean(rnx_bytes)


def make_nav(txt):
    return txt.replace(b'OBSERVATION', b'NAVIGATION ')


@pytest.mark.parametrize(
    'input_suffix',
    ['.rnx', '.RNX', '.21n']
)
def test_decompress_non_obs(tmp_path, rnx_bytes, input_suffix):
    # prepare
    txt = make_nav(rnx_bytes)
    sample_path = tmp_path / ('sample' + input_suffix + '.gz')
    sample_path.write_bytes(gzip.compress(txt))
    # decompress
    out_path = decompress_on_disk(sample_path)
    # check
    assert out_path.exists()
    assert out_path == tmp_path / ('sample' + input_suffix)
    assert clean(out_path.read_bytes()) == clean(txt)


compress_pairs = [
    ('.rnx', 'none', '.crx'),
    ('.rnx', 'gz', '.crx.gz'),
    ('.rnx', 'bz2', '.crx.bz2'),
    ('.RNX', 'gz', '.CRX.gz'),
    ('.21o', 'bz2', '.21d.bz2'),
    ('.21O', 'gz', '.21D.gz'),
    ('.crx', 'gz', '.crx.gz'),
    ('.21d', 'gz', '.21d.gz'),
    ('.crx', 'none', '.crx'),
    ('.21d', 'none', '.21d'),
]


@pytest.mark.parametrize(
    'input_suffix, compression, expected_suffix',
    compress_pairs
)
def test_compress(tmp_path, crx_sample, rnx_bytes, input_suffix, compression, expected_suffix):
    # prepare
    in_file = 'sample' + input_suffix
    sample_path = tmp_path / in_file
    shutil.copy(get_data_path(in_file), sample_path)
    # compress
    converted = compress(sample_path, compression=compression)
    # check
    assert clean(decompress(converted)) == clean(rnx_bytes)
    converted = compress(sample_path.read_bytes(), compression=compression)
    assert clean(decompress(converted)) == clean(rnx_bytes)


@pytest.mark.parametrize(
    'input_suffix, compression, expected_suffix',
    compress_pairs
)
def test_compress_on_disk(tmp_path, crx_sample, rnx_bytes, input_suffix, compression,
                          expected_suffix):
    # prepare
    sample_path = tmp_path / ('sample' + input_suffix)
    in_file = 'sample' + input_suffix
    shutil.copy(get_data_path(in_file), sample_path)
    # compress
    out_path = compress_on_disk(sample_path, compression=compression)
    # check
    assert out_path.exists()
    assert out_path == tmp_path / ('sample' + expected_suffix)
    assert clean(decompress(out_path)) == clean(rnx_bytes)


@pytest.mark.parametrize(
    'input_suffix',
    ['.rnx', '.RNX', '.21n']
)
def test_compress_non_obs(tmp_path, rnx_bytes, input_suffix):
    # prepare
    txt = make_nav(rnx_bytes)
    sample_path = tmp_path / ('sample' + input_suffix)
    sample_path.write_bytes(txt)
    # compress
    out_path = compress_on_disk(sample_path)
    # check
    assert out_path.exists()
    assert out_path == tmp_path / ('sample' + input_suffix + '.gz')
    assert clean(decompress(out_path)) == clean(txt)


def test_invalid_input(crx_str, rnx_bytes):
    with pytest.raises(ValueError):
        decompress(io.BytesIO(rnx_bytes))
    with pytest.raises(ValueError):
        compress(io.BytesIO(rnx_bytes))


@pytest.mark.parametrize(
    'input_name',
    ['sample', 'sample.crxx', 'sample.abo', 'sample.abd', 'sample.21dd']
)
def test_invalid_name(tmp_path, rnx_sample, input_name):
    sample_path = tmp_path / input_name
    shutil.copy(rnx_sample, sample_path)
    with pytest.raises(ValueError) as excinfo:
        decompress_on_disk(sample_path)
    msg = excinfo.value.args[0]
    assert msg.endswith('is not a valid RINEX file name')
    with pytest.raises(ValueError) as excinfo:
        compress_on_disk(sample_path)
    msg = excinfo.value.args[0]
    assert msg.endswith('is not a valid RINEX file name')



import argparse
import sys
import warnings
from contextlib import contextmanager
from pathlib import Path
from typing import List, Optional

from hatanaka import __version__, compress, compress_on_disk, decompress, decompress_on_disk, \
    rnxcmp_version

__all__ = ['decompress_cli', 'compress_cli']


def decompress_cli(args: Optional[List[str]] = None) -> int:
    if args is None:
        args = sys.argv[1:]

    parser = argparse.ArgumentParser(
        description='Decompress compressed RINEX files.',
        epilog='This program will decompress any RINEX files compressed with Hatanaka compression'
               '(.crx|.##d) and/or with a conventional compression format (.gz|.Z|.zip|.bz2) to '
               'their plain RINEX counterpart. Already decompressed files are ignored. '
               'Exit codes: 0 - success, 1 - error, 2 - warning.'
    )
    parser.add_argument('files', type=Path, nargs='*',
                        help='Compressed RINEX files. '
                             'stdin and stdout are used if no input files are provided.')
    parser.add_argument(
        '-s', '--skip-strange-epochs', action='store_true',
        help='This option may be used for salvaging usable data when middle of the'
             'Compact RINEX file is missing. The data after the missing part, are, however, '
             'useless until the compression operation of all data are initialized at some epoch. '
             'Combination with use of -e option of RNX2CRX may be effective.\n'
             'Caution: It is assumed that no change in the list of data types '
             'happens in the lost part of the data.')
    _add_common_args(parser)
    args = parser.parse_args(args)
    return _run(decompress, decompress_on_disk, args, skip_strange_epochs=args.skip_strange_epochs)


def compress_cli(args: Optional[List[str]] = None) -> int:
    if args is None:
        args = sys.argv[1:]

    parser = argparse.ArgumentParser(
        description='Compress RINEX files.',
        epilog='This program applies Hatanaka and optionally a conventional '
               'compression (gzip by default) to any provided RINEX files. '
               'Already compressed files are ignored. '
               'Exit codes: 0 - success, 1 - error, 2 - warning.'
    )
    parser.add_argument('files', type=Path, nargs='*',
                        help='RINEX files. '
                             'stdin and stdout are used if no input files are provided.')
    parser.add_argument('-c', '--compression', default='gz', choices=['gz', 'bz2', 'none'],
                        help='which compression to apply in addition to Hatanaka compression '
                             '(default: gz)')
    parser.add_argument(
        '-s', '--skip-strange-epochs', action='store_true',
        help='warn and skip strange epochs instead of raising an exception')
    parser.add_argument(
        '-e', '--reinit-every-nth', type=int, metavar='#',
        help='Initialize the compression operation at every # epochs. '
             'When some part of the Compact RINEX file is lost, the data can not be recovered '
             'thereafter until all the data arc are initialized for differential operation. '
             'This option may be used to increase chances to recover parts of data by using '
             'the --skip-strange-epochs option of rinex-decompress at the cost of '
             'increasing the file size.')
    _add_common_args(parser)
    args = parser.parse_args(args)
    return _run(compress, compress_on_disk, args,
                compression=args.compression,
                skip_strange_epochs=args.skip_strange_epochs,
                reinit_every_nth=args.reinit_every_nth)


def _run(func, func_on_disk, args, **kwargs):
    missing_files = [x for x in args.files if not x.exists()]
    if missing_files:
        for f in missing_files:
            print(f"Error: '{str(f)}' was not found", file=sys.stderr)
            exit(1)

    for in_file in args.files:
        with _record_warnings() as warning_list:
            out_file = func_on_disk(in_file, **kwargs)
        if out_file == in_file:
            print(f'{str(in_file)} is already {func.__name__}ed')
        else:
            print(f'Created {str(out_file)}')
        assert out_file.exists()
        if args.delete:
            if len(warning_list) == 0 and in_file != out_file:
                in_file.unlink()
                print(f'Deleted {str(in_file)}')

    if len(args.files) == 0:
        with _record_warnings() as warning_list:
            converted = func(sys.stdin.buffer.read(), **kwargs)
            sys.stdout.buffer.write(converted)

    if len(warning_list) > 0:
        return 2
    return 0


def _add_common_args(parser):
    parser.add_argument('-d', '--delete', action='store_true',
                        help='delete the input file if conversion '
                             'finishes without any errors and warnings')
    parser.add_argument('--version', action='version', version=__version__)
    parser.add_argument('--rnxcmp-version', action='version', version=rnxcmp_version)


@contextmanager
def _record_warnings():
    with warnings.catch_warnings(record=True) as warning_list:
        yield warning_list
    for w in warning_list:
        warnings.showwarning(message=w.message, category=w.category, filename=w.filename,
                             lineno=w.lineno, file=w.file, line=w.line)



import pytest

from hatanaka import HatanakaException, crx2rnx, rnx2crx
from .conftest import clean


def test_rnx2crx_str(rnx_str, crx_str):
    assert clean(rnx2crx(rnx_str)) == clean(crx_str)


def test_rnx2crx_bytes(rnx_bytes, crx_bytes):
    assert clean(rnx2crx(rnx_bytes)) == clean(crx_bytes)


def test_crx2rnx_str(crx_str, rnx_str):
    assert crx2rnx(crx_str) == rnx_str


def test_crx2rnx_bytes(crx_bytes, rnx_bytes):
    assert clean(crx2rnx(crx_bytes)) == clean(rnx_bytes)


def test_rnx2crx_str_stream(rnx_str_stream, crx_str):
    assert clean(rnx2crx(rnx_str_stream)) == clean(crx_str)


def test_rnx2crx_bytes_stream(rnx_bytes_stream, crx_bytes):
    assert clean(rnx2crx(rnx_bytes_stream)) == clean(crx_bytes)


def test_crx2rnx_str_stream(crx_str_stream, rnx_str):
    assert crx2rnx(crx_str_stream) == rnx_str


def test_crx2rnx_bytes_stream(crx_bytes_stream, rnx_bytes):
    assert clean(crx2rnx(crx_bytes_stream)) == clean(rnx_bytes)


def test_invalid():
    with pytest.raises(HatanakaException) as excinfo:
        crx2rnx('blah')
    msg = excinfo.value.args[0]
    assert msg.startswith('The file seems to be truncated in the middle.')


def test_warning(rnx_bytes, crx_bytes):
    rnx_bytes += b'\0\0\0'
    with pytest.warns(UserWarning) as record:
        converted = rnx2crx(rnx_bytes)
    assert clean(converted) == clean(crx_bytes)
    assert len(record) == 1
    assert record[0].message.args[0].startswith('rnx2crx: null characters')


def test_rnx2crx_extra_args_good(rnx_str, crx_str):
    converted = rnx2crx(rnx_str, reinit_every_nth=1, skip_strange_epochs=True)
    assert clean(converted) == clean(crx_str)


def test_rnx2crx_extra_args_warning(rnx_str, crx_str):
    rnx_str = rnx_str.replace('R19 129262004.57708', 'G13 130321269.80108')
    with pytest.warns(UserWarning) as record:
        converted = rnx2crx(rnx_str, reinit_every_nth=1, skip_strange_epochs=True)
    assert len(record) == 1
    assert (record[0].message.args[0] ==
            'rnx2crx: Duplicated satellite in one epoch at line 15. ... skip')
    # Only the header remains
    assert clean(crx_str).startswith(clean(converted))


def test_crx2rnx_extra_args_good(rnx_str, crx_str):
    converted = crx2rnx(crx_str, skip_strange_epochs=True)
    assert clean(converted) == clean(rnx_str)


def test_non_ascii_is_tolerated(rnx_bytes, crx_bytes):
    def add_non_ascii(txt):
        return txt.replace(
            b'VERSION / TYPE',
            b'VERSION / TYPE' + 'õäü'.encode().ljust(60) + b'COMMENT\n')

    converted = crx2rnx(add_non_ascii(crx_bytes))
    assert clean(converted) == clean(add_non_ascii(rnx_bytes))
    converted = rnx2crx(add_non_ascii(rnx_bytes))
    assert clean(converted) == clean(add_non_ascii(crx_bytes))


def test_stderr_bad_encoding(crx_bytes):
    """The errors and warnings can include cited garbage from invalid files
    and we should handle that gracefully."""
    with pytest.raises(HatanakaException) as excinfo:
        crx2rnx(crx_bytes[:-1] + b'\xFF\0')
    msg = excinfo.value.args[0]
    assert msg.startswith('The file seems to be truncated')
    assert msg.endswith('\\xff<end')


if __name__ == '__main__':
    pytest.main()
